import paco.models.applications
import paco.models.networks
import paco.models.loader
import os
import troposphere
import unittest
from paco.models import load_project_from_yaml, schemas
from paco.models.project import Project

def fixtures_path():
    # find the project root directory
    # this works for pytest run from VS Code
    parts = []
    for part in os.path.abspath(os.path.dirname(__file__)).split(os.sep):
        if part == 'src':
            parts.append('fixtures')
            break
        parts.append(part)
    path = os.sep.join(parts)
    return path

def cwd_to_fixtures():
    fpath = fixtures_path()
    os.chdir(fpath)
    return fpath

class BaseTestModelLoader(unittest.TestCase):
    project_name = 'config_me'

    @classmethod
    def setUpClass(cls):
        # set the fixtures dir
        cls.path = fixtures_path()
        cls.project = load_project_from_yaml(cls.path + os.sep + cls.project_name)

class Testpacodemo(BaseTestModelLoader):

    project_name = 'pacodemo'

    def setUp(self):
        self.demo_env = self.project['netenv']['pacodemo']['demo']['us-west-2']
        self.demo_app = self.demo_env['applications']['app']

    def test_project(self):
        assert isinstance(self.project, Project)
        assert self.project.name == 'waterbear-networks'
        assert self.project.title.startswith('Waterbear Networks')

    def test_network_az(self):
        network = self.project['netenv']['pacodemo']['demo']['us-west-2'].network
        assert network.availability_zones == 2

    def test_environment_state(self):
        dev_alb_domain = self.project['netenv']['pacodemo']['dev']['us-west-2']['applications']['app'].groups['site'].resources['alb'].dns[0].domain_name
        dev_cert_domain = self.project['netenv']['pacodemo']['dev']['us-west-2']['applications']['app'].groups['site'].resources['cert'].domain_name
        assert dev_alb_domain == "dev.pacodemo.waterbear.cloud"
        assert dev_cert_domain == "dev.pacodemo.waterbear.cloud"

    def test_deployment_pipeline(self):
        deploy_pipeline = self.project['netenv']['pacodemo']['demo']['us-west-2']['applications']['app'].groups['cicd'].resources['pipeline']
        assert schemas.IDeploymentPipeline.providedBy(deploy_pipeline)
        assert deploy_pipeline.stages['source']['github'].type, 'GitHub.Source'
        ecrpipe = self.project['netenv']['pacodemo']['demo']['us-west-2']['applications']['app'].groups['cicd'].resources['ecrpipe']
        assert ecrpipe.source['ecr'].type, 'ECR.Source'

        # Test CodeBuild
        codebuild = deploy_pipeline.stages['build']['codebuild']
        ecs = codebuild.release_phase.ecs
        assert schemas.IDeploymentPipelineBuildReleasePhase(codebuild.release_phase)
        assert ecs[0].command, 'docker rake:deploy'

    def test_ne_vpc(self):
        vpc = self.project['netenv']['pacodemo']['demo']['us-west-2'].network.vpc
        assert vpc.enable_dns_support == True
        assert vpc.vpn_gateway['app'].enabled == False
        assert vpc.private_hosted_zone.name == 'example.internal'
        assert vpc.segments['database'].enabled == True
        lb_sg = vpc.security_groups['app']['lb']
        assert lb_sg.ingress[0].from_port == 443
        assert lb_sg.ingress[0].name == 'HTTPS'
        assert lb_sg.egress[0].name == 'ANY'

    def test_env_network_merged(self):
        demo_env = self.project['netenv']['pacodemo']['demo']
        assert demo_env['us-west-2'].network.availability_zones == 2
        assert isinstance( demo_env['us-west-2']['network'].vpc, paco.models.networks.VPC)
        assert demo_env['default'].network.vpc.cidr == '10.0.0.0/16'
        assert demo_env['default'].network.vpc.segments['webapp'].az1_cidr == '10.0.3.0/24'
        assert demo_env['default'].network.vpc.segments['webapp'].internet_access == False
        assert demo_env['default'].network.vpc.segments['public'].internet_access == True

    def test_lbapplication(self):
        demo_env = self.project['netenv']['pacodemo']['demo']['us-west-2']
        alb = demo_env['applications']['app'].groups['site'].resources['alb']
        assert schemas.ILoadBalancer.providedBy(alb)
        assert schemas.IListener.providedBy(alb.listeners['https'])
        assert alb.listeners['https'].rules['app_forward'].name == 'app_forward'
        assert alb.listeners['https'].rules['app_forward'].rule_type == 'forward'

    def test_lbnetwork(self):
        demo_env = self.project['netenv']['pacodemo']['demo']['us-west-2']
        nlb = demo_env['applications']['app'].groups['site'].resources['nlb']
        assert schemas.INetworkLoadBalancer.providedBy(nlb)
        # assert schemas.IListener.providedBy(alb.listeners['https'])
        # assert alb.listeners['https'].rules['app_forward'].name == 'app_forward'
        # assert alb.listeners['https'].rules['app_forward'].rule_type == 'forward'


    def test_asg(self):
        demo_env = self.project['netenv']['pacodemo']['demo']['us-west-2']
        asg = demo_env.applications['app'].groups['site'].resources['webapp']

        # rolling_udpate_policy
        assert asg.rolling_update_policy.max_batch_size == 1
        assert asg.rolling_update_policy.min_instances_in_service == 0
        assert asg.rolling_update_policy.pause_time == 'PT1M'
        assert asg.rolling_update_policy.wait_on_resource_signals == False

        # CloudFormation Init
        cfn_init = asg.cfn_init
        assert schemas.ICloudFormationInit.providedBy(cfn_init)
        assert cfn_init.raw_config['config_sets']['ascending'][0] == 'config1'
        configurations = cfn_init.configurations
        assert schemas.ICloudFormationConfigurations.providedBy(configurations)
        assert configurations['config1'].packages.rubygems['chef'][0] == "0.10.2"
        assert configurations['config1'].commands['test'].command == "ls"
        assert configurations['config1'].commands['test'].cwd == "~"
        assert configurations['config1'].commands['test'].env['CFNTEST'] ==  "I come from config1."

        config_sets = cfn_init.config_sets
        assert schemas.ICloudFormationConfigSets.providedBy(config_sets)
        assert config_sets['ascending'][0] == 'config1'

        # Services
        service = configurations['config1'].services.sysvinit['apache2']
        assert service.enabled == True
        assert service.commands[1] == 'two'
        assert service.files[0] == "/etc/cfn/cfn-hup.conf"

        # Test files - check for CloudFormation functions such as Fn::Sub or !Sub
        local_file = configurations['config1'].files['/tmp/local-file']
        assert type(local_file.content) == troposphere.Sub

        plain_file = configurations['config1'].files['/tmp/plain-file']
        assert plain_file.content == "Nothing to see here."

        #fullyaml = configurations['config1'].files['/fullyaml.txt'].content
        mysql_file = configurations['config1'].files['/tmp/setup.mysql']
        assert mysql_file.mode == "000644"
        assert mysql_file.group == "root"
        #assert mysql_file.content == '!Sub\nCREATE DATABASE ${DBName};\n'

        tag_file = configurations['config1'].files['/tag.txt']
        assert tag_file.content == '!Join\n- \'\'\n- - "#!/bin/bash\\n"\n  - !Ref: AWS::StackName\n!Ref: SomeLogicalId\n'

        # Use expressions such Fn::Sub
        #test_file = configurations['config1'].files['/test-file.txt']
        #assert test_file.content == 'Fn::Sub | TEST FILE ${DBName};'

        # Test cfn_init Parameters
        assert cfn_init.parameters['TestString'] == 'catdog'
        assert cfn_init.parameters['TestNumber'] == 10

        # Test IAM Role
        assert asg.instance_iam_role.enabled == True
        assert asg.instance_iam_role.name == 'instance_iam_role'

        norole_asg = demo_env.applications['app'].groups['site'].resources['norole']
        assert norole_asg.instance_iam_role.enabled == False

        # SSH Access
        assert asg.ssh_access.users, ['bdobbs']
        assert asg.ssh_access.groups, ['developers']

        # ECS
        ecs_asg = demo_env.applications['app'].groups['container'].resources['ecs_asg']
        assert ecs_asg.ecs.cluster, 'paco.ref netenv.pacodemo.demo.us-west-2.applications.app.groups.container.resources.ecs_cluster'


    def test_netenv_refs(self):
        demo_env = self.project['netenv']['pacodemo']['demo']['us-west-2']

        # netenf.ref in a List
        ref_value = demo_env.applications['app'].groups['site'].resources['alb'].security_groups[0]
        assert ref_value == "paco.ref netenv.pacodemo.demo.us-west-2.network.vpc.security_groups.app.lb"

    def test_get_all_nodes(self):
        # ToDo: EnvironmentRegion.applications is seen twice because it's both
        # an attribute and a dict key - refactor so it's only attribute
        # but it's breaking change ...
        return
        # check to ensure each node is only visited once
        seen = {}
        for node in paco.models.loader.get_all_nodes(self.project):
            node_id = id(node)
            if node_id in seen:
                # core metric objects are the same
                name = getattr(node, 'name', '')
                if name not in ('ec2core_builtin_metric'):
                    self.fail("Node seen a second time")
            else:
                seen[node_id] = node

    def test_env_override(self):
        dev_env = self.project['netenv']['pacodemo']['dev']['us-west-2']
        bastion = dev_env['applications']['app'].groups['bastion'].resources['instance']
        assert bastion.desired_capacity == 0
        #bastion = dev_env['applications']['app']['groups']['bastion']['resources']['bastion']
        #assert bastion['instance_key_pair'] == 'wbsites-dev-us-west-2'
        #assert bastion['desired_capacity'] == 0

    def test_alarms(self):
        dev_env = self.project['netenv']['pacodemo']['dev']['us-west-2']
        dev_bastion = dev_env['applications']['app'].groups['bastion'].resources['instance']
        demo_env = self.project['netenv']['pacodemo']['demo']['us-west-2']
        demo_bastion = demo_env['applications']['app'].groups['bastion'].resources['instance']
        demo_webapp = demo_env['applications']['app'].groups['site'].resources['webapp']

        # test typical set-up: the dev env does not have Alarms but demo env does
        assert len(dev_bastion.monitoring.alarm_sets.values()) == 0
        assert len(demo_bastion.monitoring.alarm_sets.values()) > 0
        # bastion has the ASG instance-health-core AlarmSet
        # The Loader loads each AlarmSet into an AlarmSets object
        # then it can overrides existing Alarm settings
        assert schemas.ICloudWatchAlarm.providedBy(demo_bastion.monitoring.alarm_sets['instance-health-core']['StatusCheck-Critical'])
        assert demo_bastion.monitoring.alarm_sets['instance-health-cwagent']['SwapPercent-Low'].evaluation_periods == 5
        assert schemas.ICloudWatchAlarm.providedBy(demo_webapp.monitoring.alarm_sets['launch-health']['GroupPendingInstances-Low'])
        assert demo_webapp.monitoring.alarm_sets['instance-health-cwagent']['SwapPercent-Low'].evaluation_periods == 15
        assert demo_webapp.monitoring.alarm_sets['instance-health-cwagent']['SwapPercent-Low'].threshold == 10.0

        # LogAlarm
        assert demo_webapp.monitoring.alarm_sets['log-test']['ApacheError'].type == 'LogAlarm'
        assert demo_webapp.monitoring.alarm_sets['log-test']['ApacheError'].log_set_name == 'apache'
        assert demo_webapp.monitoring.alarm_sets['log-test']['ApacheError'].log_group_name == 'error'

    def test_dbparameters(self):
        demo_env = self.project['netenv']['pacodemo']['demo']['us-west-2']
        dbparams = demo_env['applications']['app'].groups['site'].resources['dbparams']
        assert schemas.IDBParameterGroup.providedBy(dbparams)
        assert dbparams.parameters['block_encryption_mode'] == 'aes-128-ecb'

    def test_iamuser_resource(self):
        demo_env = self.project['netenv']['pacodemo']['demo']['us-west-2']
        github_user = demo_env['applications']['app'].groups['cicd'].resources['github_user']
        assert schemas.IIAMUserResource.providedBy(github_user)
        assert github_user.allows[0] == 'paco.ref netenv.pacodemo.demo.us-west-2.app.app.groups.cicd.resources.ecrrepo'

    def test_pinpointapp(self):
        pinpointapp = self.demo_app.groups['market'].resources['pinpointapp']
        assert schemas.IPinpointApplication(pinpointapp)
        assert pinpointapp.email_channel.from_address == 'bob@example.com'
        assert pinpointapp.sms_channel.sender_id == 'MisterId'

    def test_cloudwatch_logging(self):
        demo_env = self.project['netenv']['pacodemo']['demo']['us-west-2']
        demo_webapp = demo_env['applications']['app'].groups['site'].resources['webapp']
        linux_log_set = demo_webapp.monitoring.log_sets['amazon_linux']

        # Log Set
        assert schemas.ICloudWatchLogSet.providedBy(linux_log_set)
        assert linux_log_set.expire_events_after_days == '1'

        # Log Group
        assert schemas.ICloudWatchLogGroup.providedBy(linux_log_set.log_groups['audit'])
        assert linux_log_set.log_groups['audit'].log_group_name == 'puppydog'

        # Log Source
        assert schemas.ICloudWatchLogSource.providedBy(linux_log_set.log_groups['audit'].sources['audit'])
        assert linux_log_set.log_groups['audit'].sources['audit'].log_stream_name == "audit-{instance_id}"

        # metric filters
        assert linux_log_set.log_groups['audit'].metric_filters['authorization_failures'].filter_pattern.startswith('{ ($')

    def test_instantiate_resources(self):
        # Route53
        assert self.project['resource']['route53'].hosted_zones['pacodemo'].domain_name, 'pacodemo.example.com'
        # CodeCommit
        assert self.project['resource']['codecommit']['pacodemo']['app'].account, 'paco.ref accounts.data'
        # EC2
        assert self.project['resource']['ec2'].keypairs['pacodemo_dev'].account, 'paco.ref accounts.dev'

    def test_resource_account(self):
        dev_env = self.project['netenv']['pacodemo']['dev']['us-west-2']
        bastion = dev_env['applications']['app'].groups['bastion'].resources['instance']
        account = bastion.get_account()
        assert account.name, 'dev'

    def test_notifications(self):
        demo_env = self.project['netenv']['pacodemo']['demo']['us-west-2']
        demo_app = demo_env['applications']['app']

        # notifications for applications
        assert schemas.IAlarmNotifications.providedBy(demo_app.notifications)
        assert schemas.IAlarmNotification.providedBy(demo_app.notifications['team_bob'])
        assert demo_app.notifications['team_bob'].classification, 'performance'

        # test application override
        assert demo_app.notifications['team_bob'].groups[0], 'jim_is_the_new_bob'

        # notification for a resource
        webapp = demo_app.groups['site'].resources['webapp']
        notif_group = webapp.monitoring.notifications['santaclaus']
        assert notif_group.groups[0], 'santa'

        # notification for an alarm set
        alarm_set = webapp.monitoring.alarm_sets['instance-health-cwagent']
        assert schemas.IAlarmNotifications.providedBy(alarm_set.notifications)
        assert alarm_set.notifications['alarmsetnotif'].groups[0], 'misterteam'

        # notification for just a single alarm
        alarm_with_notif = alarm_set['SwapPercent-Low']
        assert schemas.IAlarmNotifications.providedBy(alarm_with_notif.notifications)
        assert alarm_with_notif.notifications['singlealarm']

    def test_alarm_notifications(self):
        demo_env = self.project['netenv']['pacodemo']['demo']['us-west-2']
        webapp = demo_env['applications']['app'].groups['site'].resources['webapp']

        alarm_set = webapp.monitoring.alarm_sets['instance-health-cwagent']

        # alarm with specific notification
        alarm_one = alarm_set['SwapPercent-Low']
        assert alarm_one.notification_groups, ['oneguygetsthis', 'misterteam', 'santa', 'jim_is_the_new_bob']

        # alarm with alarm set notification
        alarm_two = alarm_set['SwapPercent-Critical']
        assert alarm_two.notification_groups, ['misterteam', 'santa', 'jim_is_the_new_bob']

        # alarm with resource notification
        alarm_set = webapp.monitoring.alarm_sets['launch-health']
        alarm_three = alarm_set['GroupPendingInstances-Low']
        assert alarm_three.notification_groups, ['santa']

        # alarm with no notifications, is filtered out of on severity and performance
        alarm_four = alarm_set['GroupPendingInstances-Critical']
        assert len(alarm_four.notification_groups) == 0

        # alarm with only app notification
        alarm_set = webapp.monitoring.alarm_sets['instance-health-core']
        alarm_five = alarm_set['CPUTotal-Low']
        assert alarm_five.notification_groups, ['santa']

    def test_sns(self):
        sns = self.project['resource']['sns']
        assert len(sns.default_locations), 1
        assert len(sns.topics), 2

    def test_ec2(self):
        ec2 = self.project['resource']['ec2']
        assert ec2.keypairs['pacodemo_dev'].region, 'us-west-2'
        assert ec2.users['bdobbs'].full_name, 'Bob Dobbs'
        assert ec2.groups['developers'].members, ['Bob Dobbs']

    def test_ecs(self):
        demo_env = self.project['netenv']['pacodemo']['demo']['us-west-2']
        cluster = demo_env['applications']['app'].groups['container'].resources['ecs_cluster']
        assert schemas.IECSCluster.providedBy(cluster)
        assert cluster.capacity_providers[0].base, 1
        ecs_service_config = demo_env['applications']['app'].groups['container'].resources['ecs_services']

        assert ecs_service_config.setting_groups['app_pool'].secrets[0].name, 'TWO_SECRET'
        assert ecs_service_config.setting_groups['app_pool'].environment[1].value, 'down'

        simple_app_service = ecs_service_config.services['simple_app']
        assert ecs_service_config.services['simple_app'].desired_count, 2
        assert ecs_service_config.services['simple_app'].load_balancers[0].container_name, 'hello'
        hello_def = ecs_service_config.task_definitions['hello_web'].container_definitions['hello']
        assert hello_def.cpu, 10
        assert hello_def.depends_on[0].condition, 'START'
        assert hello_def.health_check.retries, 5
        assert hello_def.ulimits[0].hard_limit, 1000
        assert hello_def.user, 'www-data'
        assert hello_def.secrets[0].value_from, 'paco.ref netenv.pacodemo.secrets_manager.one.two.another'
        assert hello_def.environment[0].name, 'hello'

        assert simple_app_service.target_tracking_scaling_policies['memory'].scale_in_cooldown, 300
        assert simple_app_service.target_tracking_scaling_policies['cpu'].predefined_metric, 'ECSServiceAverageCPUUtilization'
        assert simple_app_service.target_tracking_scaling_policies['cpu'].target, 70

        # Capacity Providers
        assert simple_app_service.capacity_providers[0].base, 1
        assert simple_app_service.capacity_providers[0].weight, 1

    def test_cognito(self):
        auth_app = self.demo_app.groups['auth']
        userpool = auth_app.resources['userpool']
        assert schemas.ICognitoUserPool(userpool)
        assert userpool.app_clients['bob'].generate_secret, True
        assert userpool.user_creation.invite_message_templates.email_message, "Would you like to test the email?"
        assert userpool.lambda_triggers.pre_sign_up.startswith('paco.ref ')
        idp = auth_app.resources['ident']
        assert schemas.ICognitoIdentityPool(idp)

    def test_rds(self):
        demo_env = self.project['netenv']['pacodemo']['demo']['us-west-2']
        pg_aurora = demo_env['applications']['app'].groups['site'].resources['pg_aurora']
        assert pg_aurora.db_instances['first'].db_instance_type, 'db.t3.small'
        assert pg_aurora.default_instance.db_instance_type, 'db.t3.medium'
        assert pg_aurora.enable_kms_encryption, True
        assert pg_aurora.cluster_event_notifications.groups, ['bob']
        assert pg_aurora.default_instance.event_notifications.event_categories[0], 'availability'
        assert pg_aurora.db_instances['first'].event_notifications.event_categories[0], 'failure'

    def test_lambda(self):
        demo_env = self.project['netenv']['pacodemo']['demo']['us-west-2']
        lmbda = demo_env['applications']['notification'].groups['lambda'].resources['function']
        assert schemas.ILambda.providedBy(lmbda)
        assert lmbda.handler, 'notification.lambda_handler'
        assert lmbda.memory_size, 128
        assert len(lmbda.layers), 1

    def test_esdomain(self):
        demo_env = self.project['netenv']['pacodemo']['demo']['us-west-2']
        esdomain = demo_env['applications']['app'].groups['site'].resources['esdomain']
        assert schemas.IElasticsearchDomain.providedBy(esdomain)

    def test_paco_project_version(self):
        # test that a version loaded ... we will fiddle with this number in fixtures
        # as we update paco.models
        assert len(self.project.paco_project_version) > 2

    def test_cloudtrail(self):
        cloudtrail = self.project['resource']['cloudtrail']
        assert schemas.ICloudTrailResource.providedBy(cloudtrail)
        trail = cloudtrail.trails['basic_trail']
        assert schemas.ICloudTrail.providedBy(trail)
        assert trail.enable_log_file_validation == True
        assert trail.cloudwatchlogs_log_group.log_group_name, 'CloudTrail'
        assert trail.cloudwatchlogs_log_group.expire_events_after_days, '14'

    def test_config(self):
        config = self.project['resource']['config'].config
        assert config.global_resources_region == 'us-west-2'
        assert config.delivery_frequency == 'Six_Hours'

    def test_api_gateway_rest_api(self):
        demo_env = self.project['netenv']['pacodemo']['demo']['us-west-2']
        api = demo_env['applications']['app'].groups['restapi'].resources['api_gateway_rest_api']
        assert len(api.body_file_location) > 1
        api.resources['idle'].path_part, '/idle'
        api.resources['idle'].child_resources['view'].path_part, '{view}'
        api.resources['idle'].child_resources['view'].child_resources['one'].path_part, 'one'
        api.resources['idle'].child_resources['view'].child_resources['two'].path_part, 'two'

    def test_health_checks(self):
        demo_env = self.project['netenv']['pacodemo']['demo']['us-west-2']
        health_checks = demo_env['applications']['app'].monitoring.health_checks
        assert schemas.IHealthChecks.providedBy(health_checks)
        pinger = health_checks['external_ping']
        assert pinger.match_string, 'alive!'
        assert pinger.port, 80

    def test_multiple_apps_one_netenv(self):
        demo_env = self.project['netenv']['pacodemo']['demo']['us-west-2']
        apps = demo_env['applications']
        assert schemas.IApplication.providedBy(apps['appmouse'])
        assert schemas.IApplication.providedBy(apps['appelephant'])
        assert schemas.IApplication.providedBy(apps['app'])

    def test_codedeploy_application(self):
        demo_env = self.project['netenv']['pacodemo']['demo']['us-west-2']
        codedeploy = demo_env['applications']['app'].groups['cicd'].resources['codedeploy']
        assert schemas.ICodeDeployApplication.providedBy(codedeploy)
        assert codedeploy.compute_platform == "Server"
        assert len(codedeploy.deployment_groups['deployment'].autoscalinggroups) == 1
        assert codedeploy.deployment_groups['deployment'].revision_location_s3.bundle_type == 'zip'

    def test_backup(self):
        demo_env = self.project['netenv']['pacodemo']['demo']['us-west-2']
        vaults = demo_env.backup_vaults
        assert schemas.IBackupVaults.providedBy(vaults)
        myvault = vaults['myapp']
        assert myvault.title == "All data for MyApp (myapp) application"
        assert myvault.plans['ebs_daily'].plan_rules[0].schedule_expression == 'cron(0 7 ? * * *)'

        assert myvault.plans['ebs_daily'].plan_rules[0].copy_actions[0].destination_vault.startswith('arn:aws:backup')
        assert myvault.plans['ebs_daily'].plan_rules[0].copy_actions[1].lifecycle_delete_after_days == 100

    # dashboard_file is not reading correctly? disable for now
    def test_dashboard(self):
        demo_env = self.project['netenv']['pacodemo']['demo']['us-west-2']
        dashboard = demo_env['applications']['app'].groups['site'].resources['dashboard']
        assert schemas.ICloudWatchDashboard.providedBy(dashboard)
        assert dashboard.title == "Demo-Dashboard"
        assert 'WebAsg.name' in dashboard.variables

    def test_dynamodb(self):
        dynamodb = self.demo_app.groups['storage'].resources['dynamodb']
        assert schemas.IDynamoDB.providedBy(dynamodb)
        assert dynamodb.default_provisioned_throughput.read_capacity_units == 5
        concert = dynamodb.tables['concert']
        assert concert.attribute_definitions[0].name == 'ArtistId'
        assert concert.key_schema[0].type == 'HASH'
        assert concert.global_secondary_indexes[0].index_name == 'GSI'
        assert concert.global_secondary_indexes[0].projection.type == 'KEYS_ONLY'
        assert concert.provisioned_throughput.write_capacity_units == 10
        assert concert.target_tracking_scaling_policy.min_capacity == 5
        discography = dynamodb.tables['discography']
        assert discography.attribute_definitions[1].type == 'S'

    def test_iotcore(self):
        demo_env = self.project['netenv']['pacodemo']['demo']['us-west-2']
        iottopic = demo_env['applications']['app'].groups['iot'].resources['iottopic']
        assert schemas.IIoTTopicRule.providedBy(iottopic)
        assert iottopic.sql == "SELECT * FROM 'iot/myTestTopic'"

        iotpolicy = demo_env['applications']['app'].groups['iot'].resources['iotpolicy']
        assert schemas.IIoTPolicy.providedBy(iotpolicy)
        assert iotpolicy.variables['sensor_topic_arn'].startswith('paco.ref ')

    def test_iotanalyticspipeline(self):
        demo_env = self.project['netenv']['pacodemo']['demo']['us-west-2']
        iotpipeline = demo_env['applications']['app'].groups['iot'].resources['raw_analysis']
        assert schemas.IIoTAnalyticsPipeline.providedBy(iotpipeline)
        assert iotpipeline.channel_storage.key_prefix == 'raw_input/'
        assert iotpipeline.datastore_storage.expire_events_after_days == 30
        assert iotpipeline.pipeline_activities['extra_bit'].attributes['key1'] == 'heyguy'

        # datasets
        sample = iotpipeline.datasets['sample']
        assert sample.query_action.sql_query == "SELECT * FROM example"
        assert sample.content_delivery_rules['s3dump'].s3_destination.key == "/DataSet/!{iotanalytics:scheduleTime}/!{iotanalytics:versionId}.csv"
        assert sample.version_history == 2
        assert sample.expire_events_after_days == 3

    def test_ssm_documents(self):
        ssm_documents = self.project['resource']['ssm'].ssm_documents
        assert ssm_documents['my_ssm_doc'].locations[0].regions[0] == 'eu-central-1'
        assert ssm_documents['my_ssm_doc'].document_type == 'Command'



import os.path
import boto3
import json
import pathlib
from operator import itemgetter


def update_gen_vocabulary():
    gen_vocabulary_path = os.path.dirname(__file__) + os.sep + '..' + os.sep + 'models' + os.sep + 'gen_vocabulary.py'
    gen_vocabulary_path_new = gen_vocabulary_path + '.new'
    print("Updating paco.models generated vocabulary")
    print(f"file path: {gen_vocabulary_path}")
    print()
    with open(gen_vocabulary_path_new, 'w') as f:
        f.write("""# Generated file: do not edit
# see README.md on how to run paco_update_gen_vocabulary to update this with the latest data from AWS

from zope.schema.vocabulary import SimpleVocabulary

""")
        print("Updating EC2 AMI Ids")
        regions = [
            'us-east-2',
            'us-east-1',
            'us-west-1',
            'us-west-2',
            'ca-central-1',
            'eu-central-1',
            'eu-west-1',
            'eu-west-2',
            'eu-west-3',
            'ap-southeast-1',
            'ap-southeast-2',
        ]
        ami_types = [
            {
                'ami_yaml_name': 'amazon-linux-2-ecs',
                'ami_description': "Amazon Linux AMI 2*",
                'ami_name': "amzn2-ami-ecs-hvm-*"
            },
            {
                'ami_yaml_name': 'amazon-linux-2',
                'ami_description': "Amazon Linux 2 AMI*",
                'ami_name': "amzn2-ami-hvm-*"
            },
            {
                'ami_yaml_name': 'amazon-linux-nat',
                'ami_description': "Amazon Linux AMI*",
                'ami_name': "amzn-ami-vpc-nat-hvm-*"
            },
        ]
        ami_images = {}
        for region in regions:
            print(f"Getting AMI Ids for {region}")
            ami_images[region] = {}
            ec2_client = boto3.client('ec2', region_name=region)
            for ami_type in ami_types:
                filters = [ {
                    'Name': 'name',
                    'Values': [ami_type['ami_name']]
                },{
                    'Name': 'description',
                    'Values': [ami_type['ami_description']]
                },{
                    'Name': 'architecture',
                    'Values': ['x86_64']
                },{
                    'Name': 'owner-alias',
                    'Values': ['amazon']
                },{
                    'Name': 'owner-id',
                    'Values': ['*']
                },{
                    'Name': 'state',
                    'Values': ['available']
                },{
                    'Name': 'root-device-type',
                    'Values': ['ebs']
                },{
                    'Name': 'virtualization-type',
                    'Values': ['hvm']
                },{
                    'Name': 'hypervisor',
                    'Values': ['xen']
                },{
                    'Name': 'image-type',
                    'Values': ['machine']
                } ]
                ami_list = ec2_client.describe_images(
                    Filters=filters,
                    Owners=[
                        'amazon'
                    ]
                )
                # sorted by newest first
                image_brief = []
                for image_info in sorted(ami_list['Images'], key=itemgetter('CreationDate'), reverse=True):
                    if image_info['State'] == 'available':
                        image_brief.append({
                            'ImageId': image_info['ImageId'],
                            'CreationDate': image_info['CreationDate'],
                        })
                ami_images[region][ami_type['ami_yaml_name']] = image_brief
        f.write("ami_ids = ")
        f.write(json.dumps(ami_images, indent=2))
        f.write("\n\n")

        print("Updating IAM Managed Policies vocabulary")
        client = boto3.client('iam')
        f.write("iam_managed_policies = SimpleVocabulary.fromValues([\n")
        paginator = client.get_paginator('list_policies')
        response_iterator = paginator.paginate(Scope='AWS')
        for response in response_iterator:
            for policy in response['Policies']:
                f.write("    '{}',\n".format(policy['PolicyName']))
        f.write("])\n\n")

        print("Updating RDS Engine Verions vocabulary")
        client = boto3.client('rds', region_name='us-west-2')
        paginator = client.get_paginator('describe_db_engine_versions')
        rds_engine_versions = {}
        for engine in ['postgres','mysql', 'aurora', 'aurora-mysql', 'aurora-postgresql', 'sqlserver-ex']:
            rds_engine_versions[engine] = {}
            response_iterator = paginator.paginate(
                Engine=engine
            )
            for response in response_iterator:
                for version in response['DBEngineVersions']:
                    rds_engine_versions[engine][version['EngineVersion']] = {}
                    rds_engine_versions[engine][version['EngineVersion']]['param_group_family'] = version['DBParameterGroupFamily']
        f.write("rds_engine_versions = ")
        f.write(json.dumps(rds_engine_versions, indent=2))
        f.write("\n\n")

    pathlib.Path(gen_vocabulary_path_new).rename(gen_vocabulary_path)
    print("Done!")



"""
All things network.
"""

import paco.models.applications
import paco.models.iam
from paco.models.base import Parent, Name, Named, Deployable, ImportFrom
from paco.models import schemas
from paco.models import vocabulary
from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty
from paco.models import loader
from paco.models import references
from paco.models.resources import SSMDocuments


@implementer(schemas.INetworkEnvironments)
class NetworkEnvironments(Named, dict):
    pass

@implementer(schemas.INetworkEnvironment)
class NetworkEnvironment(Named, Deployable, dict):
    """NetworkEnvironment"""

    @property
    def environments(self):
        return self

    def resolve_ref(self, ref):
        return self


@implementer(schemas.IEnvironment)
class Environment(Named, dict):

    @property
    def env_regions(self):
        """Filter out the default and only return EnvironmentRegion instances"""
        results = {}
        for k, v in self.items():
            if k != 'default':
                results[k] = v
        return results


@implementer(schemas.IEnvironmentDefault)
class EnvironmentDefault(Named, dict):

    def __init__(self, name, __parent__):
        super().__init__(name, __parent__)

        # applications
        self.applications = paco.models.applications.ApplicationEngines('applications', self)
        self.applications.title = 'Applications'
        self.__setitem__('applications', self.applications)

        # network
        self.network = Network('network', self)
        self.network.title = 'Network'
        self.__setitem__('network', self.network)

        # secrets_manager
        self.secrets_manager = paco.models.applications.SecretsManager('secrets_manager', self)
        self.secrets_manager.title = 'SecretsManager'
        self.__setitem__('secrets_manager', self.secrets_manager)

        # backup_vaults
        self.backup_vaults = paco.models.backup.BackupVaults('backup_vaults', self)
        self.backup_vaults.title = 'Backup Vaults'
        self.__setitem__('backup_vaults', self.backup_vaults)


@implementer(schemas.IEnvironmentRegion)
class EnvironmentRegion(EnvironmentDefault, Deployable, dict, ImportFrom):

    @property
    def title_or_name(self):
        return self.region_full_name

    @property
    def region(self):
        # an EnvironmentRegion *must* always be a valid region name
        return self.__name__

    @property
    def region_full_name(self):
        return vocabulary.aws_regions[self.__name__]['full_name']

    def has_ec2lm_resources(self):
        "True if there are resources which depend upon EC2LM SSM Documents."
        for app in self['applications'].values():
            for group in app.groups.values():
                for resource in group.resources.values():
                    if schemas.IASG.providedBy(resource) and resource.is_enabled():
                        if resource.launch_options.ssm_agent == True:
                            return True
        return False


@implementer(schemas.INetwork)
class Network(Named, Deployable, dict):
    availability_zones = FieldProperty(schemas.INetwork["availability_zones"])
    aws_account = FieldProperty(schemas.INetwork["aws_account"])
    vpc = FieldProperty(schemas.INetwork["vpc"])

    def resolve_ref(self, ref):
        if ref.resource_ref == 'aws_account':
            return self.aws_account
        elif ref.resource_ref == 'availability_zones':
            return self.availability_zones
        return self

@implementer(schemas.IVPCPeeringRoute)
class VPCPeeringRoute(Parent):
    local_segment = FieldProperty(schemas.IVPCPeeringRoute["local_segment"])
    remote_segment = FieldProperty(schemas.IVPCPeeringRoute["remote_segment"])
    cidr = FieldProperty(schemas.IVPCPeeringRoute["cidr"])

@implementer(schemas.IVPCPeering)
class VPCPeering(Named, Deployable):
    "VPC Peering"
    peer_type = FieldProperty(schemas.IVPCPeering["peer_type"])
    peer_role_name = FieldProperty(schemas.IVPCPeering["peer_role_name"])
    peer_vpcid = FieldProperty(schemas.IVPCPeering["peer_vpcid"])
    peer_account_id = FieldProperty(schemas.IVPCPeering["peer_account_id"])
    peer_region = FieldProperty(schemas.IVPCPeering["peer_region"])
    network_environment  = FieldProperty(schemas.IVPCPeering["network_environment"])

@implementer(schemas.INATGateways)
class NATGateways(Named, dict):
    pass

@implementer(schemas.IVPNGateways)
class VPNGateways(Named, dict):
    pass

@implementer(schemas.ISecurityGroups)
class SecurityGroups(Named, dict):
    pass

@implementer(schemas.ISecurityGroupSets)
class SecurityGroupSets(Named, dict):
    pass

@implementer(schemas.IVPCEndpoints)
class VPCEndpoints(Named, dict):
    pass

@implementer(schemas.ISegments)
class Segments(Named, dict):
    pass

@implementer(schemas.IVPCPeerings)
class VPCPeerings(Named, dict):
    pass

@implementer(schemas.IVPC)
class VPC(Named, Deployable):
    "VPC"
    def __init__(self, name, __parent__):
        super().__init__(name, __parent__)
        self.nat_gateway = NATGateways('nat_gateway', self)
        self.vpn_gateway = VPNGateways('vpn_gateway', self)
        self.security_groups = SecurityGroupSets('security_groups', self)
        self.segments = Segments('segments', self)
        self.peering = VPCPeerings('peering', self)

    cidr = FieldProperty(schemas.IVPC["cidr"])
    enable_dns_hostnames = FieldProperty(schemas.IVPC["enable_dns_hostnames"])
    enable_dns_support = FieldProperty(schemas.IVPC["enable_dns_support"])
    enable_internet_gateway = FieldProperty(schemas.IVPC["enable_internet_gateway"])
    nat_gateway = FieldProperty(schemas.IVPC["nat_gateway"])
    vpn_gateway = FieldProperty(schemas.IVPC["vpn_gateway"])
    private_hosted_zone = FieldProperty(schemas.IVPC["private_hosted_zone"])
    security_groups = FieldProperty(schemas.IVPC["security_groups"])
    segments = FieldProperty(schemas.IVPC["segments"])
    peering = FieldProperty(schemas.IVPC["peering"])
    endpoints = FieldProperty(schemas.IVPC["endpoints"])

    def resolve_ref(self, ref):
        if ref.last_part == 'vpc':
            return self
        return self.resolve_ref_obj.resolve_ref(ref)

@implementer(schemas.IInternetGateway)
class InternetGateway(Deployable):
    pass

@implementer(schemas.INATGateway)
class NATGateway(Named, Deployable):
    type = FieldProperty(schemas.INATGateway["type"])
    availability_zone = FieldProperty(schemas.INATGateway["availability_zone"])
    segment = FieldProperty(schemas.INATGateway["segment"])
    default_route_segments = FieldProperty(schemas.INATGateway["default_route_segments"])
    security_groups = FieldProperty(schemas.INATGateway["security_groups"])
    ec2_key_pair = FieldProperty(schemas.INATGateway["ec2_key_pair"])
    ec2_instance_type = FieldProperty(schemas.INATGateway["ec2_instance_type"])

@implementer(schemas.IVPNGateway)
class VPNGateway(Named, Deployable):
    """VPN Gateway"""

@implementer(schemas.IPrivateHostedZone)
class PrivateHostedZone(Parent, Deployable):
    name = FieldProperty(schemas.IPrivateHostedZone["name"])
    vpc_associations = FieldProperty(schemas.IPrivateHostedZone["vpc_associations"])

    def __init__(self,  __parent__):
        super().__init__(__parent__)
        self.vpc_associations = []

    def resolve_ref(self, ref):
        if ref.last_part == 'private_hosted_zone':
            return self
        return self.resolve_ref_obj.resolve_ref(ref)

    @property
    def account(self):
        "The account ref this PrivateHostedZone belongs to"
        # get account from INetwork
        return self.__parent__.__parent__.aws_account

#@implementer(schemas.ISecurityGroups)
#class SecurityGroups(dict):
#    pass

@implementer(schemas.ISecurityGroup)
class SecurityGroup(Named, Deployable):
    group_name = FieldProperty(schemas.ISecurityGroup["group_name"])
    group_description = FieldProperty(schemas.ISecurityGroup["group_description"])
    ingress = FieldProperty(schemas.ISecurityGroup["ingress"])
    egress = FieldProperty(schemas.ISecurityGroup["egress"])

    def resolve_ref(self, ref):
        return ref.resource.resolve_ref_obj.resolve_ref(ref)


@implementer(schemas.ISecurityGroupRule)
class SecurityGroupRule():
    name = FieldProperty(schemas.ISecurityGroupRule["name"])
    cidr_ip = FieldProperty(schemas.ISecurityGroupRule["cidr_ip"])
    cidr_ip_v6 = FieldProperty(schemas.ISecurityGroupRule["cidr_ip_v6"])
    description = FieldProperty(schemas.ISecurityGroupRule["description"])
    from_port = FieldProperty(schemas.ISecurityGroupRule["from_port"])
    protocol = FieldProperty(schemas.ISecurityGroupRule["protocol"])
    to_port = FieldProperty(schemas.ISecurityGroupRule["to_port"])
    port = FieldProperty(schemas.ISecurityGroupRule["port"])

@implementer(schemas.IIngressRule)
class IngressRule(Parent, SecurityGroupRule):
    source_security_group = FieldProperty(schemas.IIngressRule["source_security_group"])
    source_security_group_owner = FieldProperty(schemas.IIngressRule["source_security_group_owner"])

@implementer(schemas.IEgressRule)
class EgressRule(Parent, SecurityGroupRule):
    destination_security_group = FieldProperty(schemas.IEgressRule["destination_security_group"])

@implementer(schemas.ISegment)
class Segment(Named, Deployable):
    internet_access = FieldProperty(schemas.ISegment["internet_access"])
    az1_cidr = FieldProperty(schemas.ISegment["az1_cidr"])
    az2_cidr = FieldProperty(schemas.ISegment["az2_cidr"])
    az3_cidr = FieldProperty(schemas.ISegment["az3_cidr"])
    az4_cidr = FieldProperty(schemas.ISegment["az4_cidr"])
    az5_cidr = FieldProperty(schemas.ISegment["az5_cidr"])
    az6_cidr = FieldProperty(schemas.ISegment["az6_cidr"])

    def resolve_ref(self, ref):
        if ref.parts[-1].endswith('_cidr'):
            return getattr(self, ref.parts[-1])
        else:
            stack = self.resolve_ref_obj.resolve_ref(ref)
            if stack == None:
                raise StackException(PacoErrorCode.Unknown)
            else:
                return stack
        return None

@implementer(schemas.IVPCEndpoint)
class VPCEndpoint(Named, Deployable):
    service = FieldProperty(schemas.IVPCEndpoint["service"])
    segments = FieldProperty(schemas.IVPCEndpoint["segments"])
    security_group = FieldProperty(schemas.IVPCEndpoint["security_group"])
    availability_zone = FieldProperty(schemas.IVPCEndpoint["availability_zone"])

    def resolve_ref(self, ref):
        if ref.resource_ref == 'service':
            return self.service
        else:
            stack = self.resolve_ref_obj.resolve_ref(ref)
            if stack == None:
                raise StackException(PacoErrorCode.Unknown)
            else:
                return stack
        return None





"""
All things Resources.
"""

import troposphere.apigateway
import troposphere.route53
from paco.models.base import Enablable, Parent, Named, CFNExport, Deployable, HasStack, Resource, ApplicationResource, AccountRegions
from paco.models.metrics import SNSTopics
from paco.models.iam import Role
from paco.models import references
from paco.models import schemas
from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty
from paco.models import loader
from paco.models.locations import get_parent_by_interface
from paco.models.applications import DNS
from paco.models.references import Reference
from paco.models import references


@implementer(schemas.ITopics)
class Topics(Named, dict):
    pass

class Computed(Named, dict):
    pass

@implementer(schemas.ISNS)
class SNS(Named, dict):
    default_locations = FieldProperty(schemas.ISNS['default_locations'])
    topics = FieldProperty(schemas.ISNS['topics'])

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.computed = Computed('computed', self)
        self.topics = Topics('topics', self)

    def resolve_ref(self, ref):
        return self.resolve_ref_obj.resolve_ref(ref)

@implementer(schemas.IApiGatewayBasePathMapping)
class ApiGatewayBasePathMapping(Parent):
    base_path = FieldProperty(schemas.IApiGatewayBasePathMapping['base_path'])
    stage = FieldProperty(schemas.IApiGatewayBasePathMapping['stage'])

@implementer(schemas.IApiGatewayDNS)
class ApiGatewayDNS(DNS):
    base_path_mappings = FieldProperty(schemas.IApiGatewayDNS['base_path_mappings'])
    ssl_certificate = FieldProperty(schemas.IApiGatewayDNS['ssl_certificate'])

@implementer(schemas.IApiGatewayMethods)
class ApiGatewayMethods(Named, dict):
    pass

@implementer(schemas.IApiGatewayMethodIntegrationResponse)
class ApiGatewayMethodIntegrationResponse(CFNExport):
    content_handling = FieldProperty(schemas.IApiGatewayMethodIntegrationResponse['content_handling'])
    response_parameters = FieldProperty(schemas.IApiGatewayMethodIntegrationResponse['response_parameters'])
    response_templates = FieldProperty(schemas.IApiGatewayMethodIntegrationResponse['response_templates'])
    selection_pattern = FieldProperty(schemas.IApiGatewayMethodIntegrationResponse['selection_pattern'])
    status_code = FieldProperty(schemas.IApiGatewayMethodIntegrationResponse['status_code'])

    troposphere_props = troposphere.apigateway.IntegrationResponse.props
    cfn_mapping = {
        "ContentHandling": 'content_handling',
        "ResponseParameters": 'response_parameters',
        "ResponseTemplates": 'response_templates',
        "SelectionPattern": 'selection_pattern',
        "StatusCode": 'status_code'
    }

@implementer(schemas.IApiGatewayMethodIntegration)
class ApiGatewayMethodIntegration(Parent, CFNExport):
    integration_http_method = FieldProperty(schemas.IApiGatewayMethodIntegration['integration_http_method'])
    integration_lambda = FieldProperty(schemas.IApiGatewayMethodIntegration['integration_lambda'])
    integration_responses = FieldProperty(schemas.IApiGatewayMethodIntegration['integration_responses'])
    integration_type = FieldProperty(schemas.IApiGatewayMethodIntegration['integration_type'])
    pass_through_behavior = FieldProperty(schemas.IApiGatewayMethodIntegration['pass_through_behavior'])
    request_parameters = FieldProperty(schemas.IApiGatewayMethodIntegration['request_parameters'])
    request_templates = FieldProperty(schemas.IApiGatewayMethodIntegration['request_templates'])
    uri = FieldProperty(schemas.IApiGatewayMethodIntegration['uri'])

    @property
    def integration_responses_cfn(self):
        responses = []
        for resp in self.integration_responses:
            responses.append(resp.cfn_export_dict)
        return responses

    troposphere_props = troposphere.apigateway.Integration.props
    cfn_mapping = {
        #"CacheKeyParameters": ([basestring], False),
        #"CacheNamespace": (basestring, False),
        #"ConnectionId": (basestring, False),
        #"ConnectionType": (basestring, False),
        #"ContentHandling": (basestring, False),
        "PassthroughBehavior": 'pass_through_behavior',
        "RequestTemplates": 'request_templates',
        #"TimeoutInMillis": (integer_range(50, 29000), False),
        "IntegrationResponses": 'integration_responses_cfn',
        "IntegrationHttpMethod": 'integration_http_method',
        "RequestParameters": 'request_parameters',
        "Type": 'integration_type',
        #"Credentials": computed in template according to AWS_PROXY Integration Type,
        #"Uri": computed in the template,
    }

@implementer(schemas.IApiGatewayCognitoAuthorizers)
class ApiGatewayCognitoAuthorizers(Named, dict):
    pass

@implementer(schemas.IApiGatewayCognitoAuthorizer)
class ApiGatewayCognitoAuthorizer(Named):
    identity_source = FieldProperty(schemas.IApiGatewayCognitoAuthorizer['identity_source'])
    user_pools = FieldProperty(schemas.IApiGatewayCognitoAuthorizer['user_pools'])

@implementer(schemas.IApiGatewayMethodMethodResponseModel)
class ApiGatewayMethodMethodResponseModel():
    content_type = FieldProperty(schemas.IApiGatewayMethodMethodResponseModel['content_type'])
    model_name = FieldProperty(schemas.IApiGatewayMethodMethodResponseModel['model_name'])

@implementer(schemas.IApiGatewayMethodMethodResponse)
class ApiGatewayMethodMethodResponse():
    status_code = FieldProperty(schemas.IApiGatewayMethodMethodResponse['status_code'])
    response_models = FieldProperty(schemas.IApiGatewayMethodMethodResponse['response_models'])
    response_parameters = FieldProperty(schemas.IApiGatewayMethodMethodResponse['response_parameters'])

@implementer(schemas.IApiGatewayMethod)
class ApiGatewayMethod(Resource):
    type = "ApiGatewayMethod"
    authorization_type = FieldProperty(schemas.IApiGatewayMethod['authorization_type'])
    authorizer = FieldProperty(schemas.IApiGatewayMethod['authorizer'])
    resource_name = FieldProperty(schemas.IApiGatewayMethod['resource_name'])
    http_method = FieldProperty(schemas.IApiGatewayMethod['http_method'])
    request_parameters = FieldProperty(schemas.IApiGatewayMethod['request_parameters'])
    method_responses = FieldProperty(schemas.IApiGatewayMethod['method_responses'])
    integration = FieldProperty(schemas.IApiGatewayMethod['integration'])

    @property
    def integration_cfn(self):
        return self.integration.cfn_export_dict

    def get_resource(self):
        resource = None
        node = get_parent_by_interface(self, schemas.IApiGatewayRestApi).resources
        for part in self.resource_name.split('.'):
            resource = node[part]
            node = resource.child_resources
        return resource

    troposphere_props = troposphere.apigateway.Method.props
    cfn_mapping = {
        #"ApiKeyRequired": (bool, False),
        #"AuthorizationScopes": ([basestring], False),
        #"AuthorizerId": (basestring, False),
        #"RequestModels": (dict, False),
        #"RequestValidatorId": (basestring, False),
        "AuthorizationType": 'authorization_type',
        "HttpMethod": 'http_method',
        "Integration": 'integration_cfn',
        #"MethodResponses": 'method_responses',
        "OperationName": 'title',
        "RequestParameters": 'request_parameters',
        # "ResourceId": computed in the template looked up via resource_id
        # "RestApiId": computed in the template
    }

@implementer(schemas.IApiGatewayModels)
class ApiGatewayModels(Named, dict):
    pass

@implementer(schemas.IApiGatewayModel)
class ApiGatewayModel(Resource):
    type = "ApiGatewayModel"
    content_type = FieldProperty(schemas.IApiGatewayModel['content_type'])
    description = FieldProperty(schemas.IApiGatewayModel['description'])
    schema = FieldProperty(schemas.IApiGatewayModel['schema'])

    troposphere_props = troposphere.apigateway.Model.props
    cfn_mapping ={
        "ContentType": 'content_type',
        "Description": 'description',
        "Name": 'name',
        "Schema": 'schema',
    }

@implementer(schemas.IApiGatewayResources)
class ApiGatewayResources(Named, dict):
    pass

@implementer(schemas.IApiGatewayResource)
class ApiGatewayResource(Named, dict):
    type = "ApiGatewayResource"
    path_part = FieldProperty(schemas.IApiGatewayResource['path_part'])
    enable_cors = FieldProperty(schemas.IApiGatewayResource['enable_cors'])
    # child_resources is not a FieldProperty as it's set during recursion

    @property
    def nested_name(self):
        parent = self
        parts = []
        while not schemas.IApiGatewayRestApi.providedBy(parent):
            parts.append(parent.name)
            parent = parent.__parent__.__parent__
        parts.reverse()
        return '.'.join(parts)

    troposphere_props = troposphere.apigateway.Resource.props
    cfn_mapping = {
        # ParentId: computed in template
        "PathPart": 'path_part',
        # "RestApiId": computed in template
    }

@implementer(schemas.IApiGatewayStages)
class ApiGatewayStages(Named, dict):
    pass

@implementer(schemas.IApiGatewayStage)
class ApiGatewayStage(Resource):
    deployment_id = FieldProperty(schemas.IApiGatewayStage['deployment_id'])
    description = FieldProperty(schemas.IApiGatewayStage['description'])
    stage_name = FieldProperty(schemas.IApiGatewayStage['stage_name'])

    troposphere_props = troposphere.apigateway.Stage.props
    cfn_mapping = {
        "Description": 'description',
        "StageName": 'stage_name',
    }

@implementer(schemas.IApiGatewayRestApi)
class ApiGatewayRestApi(ApplicationResource, HasStack):
    type = "ApiGatewayRestApi"
    api_key_source_type = FieldProperty(schemas.IApiGatewayRestApi['api_key_source_type'])
    binary_media_types = FieldProperty(schemas.IApiGatewayRestApi['binary_media_types'])
    body_file_location = FieldProperty(schemas.IApiGatewayRestApi['body_file_location'])
    body_s3_location = FieldProperty(schemas.IApiGatewayRestApi['body_s3_location'])
    clone_from = FieldProperty(schemas.IApiGatewayRestApi['clone_from'])
    cognito_authorizers = FieldProperty(schemas.IApiGatewayRestApi['cognito_authorizers'])
    description = FieldProperty(schemas.IApiGatewayRestApi['description'])
    dns = FieldProperty(schemas.IApiGatewayRestApi['dns'])
    endpoint_configuration = FieldProperty(schemas.IApiGatewayRestApi['endpoint_configuration'])
    fail_on_warnings = FieldProperty(schemas.IApiGatewayRestApi['fail_on_warnings'])
    minimum_compression_size = FieldProperty(schemas.IApiGatewayRestApi['minimum_compression_size'])
    parameters = FieldProperty(schemas.IApiGatewayRestApi['parameters'])
    policy = FieldProperty(schemas.IApiGatewayRestApi['policy'])
    methods = FieldProperty(schemas.IApiGatewayRestApi['methods'])
    models = FieldProperty(schemas.IApiGatewayRestApi['models'])
    resources = FieldProperty(schemas.IApiGatewayRestApi['resources'])
    stages = FieldProperty(schemas.IApiGatewayRestApi['stages'])
    _body = None

    def __init__(self, name, __parent__):
        super().__init__(name, __parent__)
        self.methods = ApiGatewayMethods('methods', self)
        self.resources = ApiGatewayResources('resources', self)
        self.stages = ApiGatewayStages('stages', self)

    @property
    def body(self):
        "Return populated string, either body or body_file_location"
        if self.body_file_location:
            return self.body_file_location
        return self._body

    @body.setter
    def body(self, value):
        self._body = value

    @property
    def endpoint_configuration_cfn(self):
        "Warp list with a dict with a 'Types' key for CloudFormation"
        return {'Types': self.endpoint_configuration }

    troposphere_props = troposphere.apigateway.RestApi.props
    cfn_mapping = {
        "ApiKeySourceType": 'api_key_source_type',
        "BinaryMediaTypes": 'binary_media_types',
        "CloneFrom": 'clone_from',
        "Description": 'description',
        # "Name": supplied by CloudFormation
        "Body": 'body',
        "FailOnWarnings": 'fail_on_warnings',
        "EndpointConfiguration": ("endpoint_configuration", "endpoint_configuration_cfn"),
        "MinimumCompressionSize": 'minimum_compression_size',
        "Parameters": 'parameters',
        "Policy": 'policy',
    }

@implementer(schemas.IEC2KeyPairs)
class EC2KeyPairs(Named, dict):
    pass

@implementer(schemas.IEC2KeyPair)
class EC2KeyPair(Named):
    region = FieldProperty(schemas.IEC2KeyPair['region'])
    account = FieldProperty(schemas.IEC2KeyPair['account'])
    keypair_name = FieldProperty(schemas.IEC2KeyPair['keypair_name'])

@implementer(schemas.IEC2Users)
class EC2Users(Named, dict):
    pass

@implementer(schemas.IEC2User)
class EC2User(Named):
    full_name = FieldProperty(schemas.IEC2User['full_name'])
    email = FieldProperty(schemas.IEC2User['email'])
    public_ssh_key = FieldProperty(schemas.IEC2User['public_ssh_key'])

@implementer(schemas.IEC2Groups)
class EC2Groups(Named, dict):
    pass

@implementer(schemas.IEC2Group)
class EC2Group(Named):
    members = FieldProperty(schemas.IEC2Group['members'])

@implementer(schemas.IEC2Resource)
class EC2Resource(Named):
    keypairs = FieldProperty(schemas.IEC2Resource['keypairs'])

    def resolve_ref(self, ref):
        if ref.parts[2] == 'keypairs':
            keypair_id = ref.parts[3]
            keypair_attr = ref.parts[4]
            try:
                keypair = self.keypairs[keypair_id]
            except KeyError:
                references.raise_invalid_reference(ref, self.keypairs, keypair_id)
            keypair = self.keypairs[keypair_id]
            value = getattr(keypair, keypair_attr, None)
            if value == None:
                references.raise_invalid_reference(ref, keypair, keypair_attr)
            return value

        return self.resolve_ref_obj.resolve_ref(ref)

@implementer(schemas.IS3Buckets)
class S3Buckets(Named, dict):
    pass

@implementer(schemas.IS3Resource)
class S3Resource(Named):
    buckets = FieldProperty(schemas.IS3Resource['buckets'])

    def __init__(self, name, __parent__):
        super().__init__(name, __parent__)
        self.buckets = S3Buckets('buckets', self)

    def resolve_ref(self, ref):
        return self.resolve_ref_obj.resolve_ref(ref)

@implementer(schemas.IRoute53RecordSet)
class Route53RecordSet():
    record_name = FieldProperty(schemas.IRoute53RecordSet["record_name"])
    type = FieldProperty(schemas.IRoute53RecordSet["type"])
    resource_records = FieldProperty(schemas.IRoute53RecordSet["resource_records"])
    ttl = FieldProperty(schemas.IRoute53RecordSet["ttl"])


@implementer(schemas.IRoute53HostedZoneExternalResource)
class Route53HostedZoneExternalResource(Named, Deployable):
    hosted_zone_id = FieldProperty(schemas.IRoute53HostedZoneExternalResource["hosted_zone_id"])
    nameservers = FieldProperty(schemas.IRoute53HostedZoneExternalResource["nameservers"])

@implementer(schemas.IRoute53HostedZone)
class Route53HostedZone(Named, Deployable):
    domain_name = FieldProperty(schemas.IRoute53HostedZone["domain_name"])
    account = FieldProperty(schemas.IRoute53HostedZone["account"])
    record_sets = FieldProperty(schemas.IRoute53HostedZone["record_sets"])
    parent_zone = FieldProperty(schemas.IRoute53HostedZone["parent_zone"])
    external_resource = FieldProperty(schemas.IRoute53HostedZone["external_resource"])
    private_hosted_zone = FieldProperty(schemas.IRoute53HostedZone["private_hosted_zone"])
    vpc_associations = FieldProperty(schemas.IRoute53HostedZone["vpc_associations"])

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.record_sets = []

    def has_record_sets(self):
        if self.record_sets != None and len(self.record_sets) > 0:
            return True
        return False

@implementer(schemas.IRoute53Resource)
class Route53Resource(Named):
    name = 'route53'
    hosted_zones = FieldProperty(schemas.IRoute53Resource["hosted_zones"])

    def __init__(self, name, parent, config_dict):
        super().__init__(name, parent)

        self.zones_by_account = {}
        if config_dict == None:
            return
        loader.apply_attributes_from_config(self, config_dict)

        for zone_id in self.hosted_zones.keys():
            hosted_zone = self.hosted_zones[zone_id]
            aws_account_ref = hosted_zone.account
            ref = Reference(aws_account_ref)
            account_name = ref.parts[1]
            if account_name not in self.zones_by_account:
                self.zones_by_account[account_name] = []
            self.zones_by_account[account_name].append(zone_id)

    def get_hosted_zones_account_names(self):
        return sorted(self.zones_by_account.keys())

    def get_zone_ids(self, account_name=None):
        if account_name != None:
            return self.zones_by_account[account_name]
        return sorted(self.hosted_zones.keys())

    def account_has_zone(self, account_name, zone_id):
        if zone_id in self.zones_by_account[account_name]:
            return True
        return False

    def resolve_ref(self, ref):
        return self.resolve_ref_obj.resolve_ref(ref)

@implementer(schemas.IRoute53HealthCheck)
class Route53HealthCheck(Resource):

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.health_checker_regions = []

    domain_name = FieldProperty(schemas.IRoute53HealthCheck["domain_name"])
    load_balancer = FieldProperty(schemas.IRoute53HealthCheck["load_balancer"])
    health_check_type = FieldProperty(schemas.IRoute53HealthCheck["health_check_type"])
    ip_address = FieldProperty(schemas.IRoute53HealthCheck["ip_address"])
    enable_sni = FieldProperty(schemas.IRoute53HealthCheck["enable_sni"])
    port = FieldProperty(schemas.IRoute53HealthCheck["port"])
    resource_path = FieldProperty(schemas.IRoute53HealthCheck["resource_path"])
    match_string = FieldProperty(schemas.IRoute53HealthCheck["match_string"])
    failure_threshold = FieldProperty(schemas.IRoute53HealthCheck["failure_threshold"])
    request_interval_fast = FieldProperty(schemas.IRoute53HealthCheck["request_interval_fast"])
    latency_graphs = FieldProperty(schemas.IRoute53HealthCheck["latency_graphs"])
    health_checker_regions = FieldProperty(schemas.IRoute53HealthCheck["health_checker_regions"])

    # All of the configuration is nested in the HealthCheckConfig prop
    # cfn_export_dict will be a HealthCheckConfig dict, this needs to be
    # wrapped in cfn_export_dict['HealthCheckConfig'] in the template
    troposphere_props = troposphere.route53.HealthCheckConfiguration.props
    cfn_mapping = {
        # 'AlarmIdentifier': (AlarmIdentifier, False),
        # 'ChildHealthChecks': ([basestring], False),
        'EnableSNI': 'enable_sni',
        'FailureThreshold': 'failure_threshold',
        # 'FullyQualifiedDomainName': computed in template,
        # 'HealthThreshold': ,
        # 'InsufficientDataHealthStatus': (basestring, False),
        # 'Inverted': (boolean, False),
        # 'IPAddress': computed in template,
        'MeasureLatency': 'latency_graphs',
        'Port': 'port',
        'Regions': 'health_checker_regions',
        'RequestInterval': 'request_interval_cfn',
        'ResourcePath': 'resource_path',
        'SearchString': 'match_string',
        'Type': 'type_cfn',
    }

    @property
    def request_interval_cfn(self):
        if self.request_interval_fast:
            return 10
        else:
            return 30

    @property
    def type_cfn(self):
        # ToDo: implement CLOUDWATCH_METRIC and CALCULATED
        if getattr(self, 'match_string', None) != None:
            if self.health_check_type == 'HTTP':
                return 'HTTP_STR_MATCH'
            elif self.health_check_type == 'HTTPS':
                return 'HTTPS_STR_MATCH'
        else:
            return self.health_check_type

@implementer(schemas.ICodeCommitUsers)
class CodeCommitUsers(Named, dict):
    pass

@implementer(schemas.ICodeCommitUser)
class CodeCommitUser(Named):
    username = FieldProperty(schemas.ICodeCommitUser["username"])
    public_ssh_key = FieldProperty(schemas.ICodeCommitUser["public_ssh_key"])
    permissions = FieldProperty(schemas.ICodeCommitUser["permissions"])

@implementer(schemas.ICodeCommitRepository)
class CodeCommitRepository(Named, Deployable):
    repository_name = FieldProperty(schemas.ICodeCommitRepository["repository_name"])
    account = FieldProperty(schemas.ICodeCommitRepository["account"])
    region = FieldProperty(schemas.ICodeCommitRepository["region"])
    description = FieldProperty(schemas.ICodeCommitRepository["description"])
    users = FieldProperty(schemas.ICodeCommitRepository["users"])
    external_resource = FieldProperty(schemas.ICodeCommitRepository["external_resource"])

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.users = CodeCommitUsers('users', self)

@implementer(schemas.ICodeCommitRepositoryGroup)
class CodeCommitRepositoryGroup(Named, dict):
    pass

@implementer(schemas.ICodeCommit)
class CodeCommit(Named, dict):

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.repo_by_account = {}

    def gen_repo_by_account(self):
        self.repo_by_account = {}
        for group_id in self.keys():
            group_config = self[group_id]
            for repo_id in group_config.keys():
                repo_config = group_config[repo_id]
                account_dict = {'group_id': group_id,
                                'repo_id': repo_id,
                                #'account_ref': repo_config.account,
                                'aws_region': repo_config.region,
                                'repo_config': repo_config }
                if repo_config.is_enabled() == False:
                    continue
                if repo_config.account in self.repo_by_account.keys():
                    if repo_config.region in self.repo_by_account[repo_config.account].keys():
                        self.repo_by_account[repo_config.account][repo_config.region].append(account_dict)
                    else:
                        self.repo_by_account[repo_config.account][repo_config.region] = [account_dict]
                else:
                    self.repo_by_account[repo_config.account] = {repo_config.region: [account_dict]}

    def repo_account_ids(self):
        return self.repo_by_account.keys()

    def account_region_ids(self, account_id):
        return self.repo_by_account[account_id].keys()

    def repo_list_dict(self, account_id, aws_region):
         return self.repo_by_account[account_id][aws_region]

    def resolve_ref(self, ref):
        return self.resolve_ref_obj.resolve_ref(ref)

    def repo_list(self):
        "List of all repositories"
        out = []
        for group in self.values():
            for repo in group.values():
                repo.repository_group = group
                out.append(repo)
        return out


@implementer(schemas.IConfig)
class Config(Resource):
    delivery_frequency = FieldProperty(schemas.IConfig["delivery_frequency"])
    global_resources_region = FieldProperty(schemas.IConfig["global_resources_region"])
    locations = FieldProperty(schemas.IConfig["locations"])
    s3_bucket_logs_account = FieldProperty(schemas.IConfig["s3_bucket_logs_account"])

    def __init__(self, name, __parent__):
        super().__init__(name, __parent__)
        self.locations = []

    def get_accounts(self):
        """
        Resolve the locations field for all accounts.
        If locations is empty, then all accounts are returned.
        """
        if self.locations == []:
            return project['accounts'].values()
        accounts = []
        project = get_parent_by_interface(self, schemas.IProject)
        for location in self.locations:
            account = references.get_model_obj_from_ref(location.account, project)
            accounts.append(account)
        return accounts

@implementer(schemas.IConfigResource)
class ConfigResource(Named):
    config = FieldProperty(schemas.IConfigResource["config"])

@implementer(schemas.ICloudTrail)
class CloudTrail(Resource):
    type = 'CloudTrail'
    accounts = FieldProperty(schemas.ICloudTrail["accounts"])
    enable_kms_encryption = FieldProperty(schemas.ICloudTrail["enable_kms_encryption"])
    enable_log_file_validation = FieldProperty(schemas.ICloudTrail["enable_log_file_validation"])
    include_global_service_events = FieldProperty(schemas.ICloudTrail["include_global_service_events"])
    is_multi_region_trail = FieldProperty(schemas.ICloudTrail["is_multi_region_trail"])
    region = FieldProperty(schemas.ICloudTrail["region"])
    s3_key_prefix = FieldProperty(schemas.ICloudTrail["s3_key_prefix"])

    def get_accounts(self):
        """
        Resolve the CloudTrail.accounts field to a list of IAccount objects from the model.
        If the field is empty, then all accounts are returned.
        """
        project = get_parent_by_interface(self, schemas.IProject)
        if self.accounts == []:
            accounts = project['accounts'].values()
        else:
            accounts = []
            for account_ref in self.accounts:
                account = references.get_model_obj_from_ref(account_ref, project)
                accounts.append(account)
        return accounts

@implementer(schemas.ICloudTrails)
class CloudTrails(Named, dict):
    pass

@implementer(schemas.ICloudTrailResource)
class CloudTrailResource(Named):

    def __init__(self, name, __parent__):
        super().__init__(name, __parent__)
        self.trails = CloudTrails('trails', self)

    def resolve_ref(self, ref):
        return self.resolve_ref_obj.resolve_ref(ref)


@implementer(schemas.IIAMUserProgrammaticAccess)
class IAMUserProgrammaticAccess(Enablable):
    access_key_1_version = FieldProperty(schemas.IIAMUserProgrammaticAccess['access_key_1_version'])
    access_key_2_version = FieldProperty(schemas.IIAMUserProgrammaticAccess['access_key_2_version'])

@implementer(schemas.IIAMUserPermission)
class IAMUserPermission(Named, Deployable):
    type = FieldProperty(schemas.IIAMUserPermission['type'])


@implementer(schemas.IIAMUserPermissionAdministrator)
class IAMUserPermissionAdministrator(IAMUserPermission):
    accounts = FieldProperty(schemas.IIAMUserPermissionAdministrator['accounts'])
    read_only = FieldProperty(schemas.IIAMUserPermissionAdministrator['read_only'])

@implementer(schemas.IIAMUserPermissionCodeCommitRepository)
class IAMUserPermissionCodeCommitRepository(Parent):
    codecommit = FieldProperty(schemas.IIAMUserPermissionCodeCommitRepository['codecommit'])
    permission = FieldProperty(schemas.IIAMUserPermissionCodeCommitRepository['permission'])
    console_access_enabled = FieldProperty(schemas.IIAMUserPermissionCodeCommitRepository['console_access_enabled'])
    public_ssh_key = FieldProperty(schemas.IIAMUserPermissionCodeCommitRepository['public_ssh_key'])

@implementer(schemas.IIAMUserPermissionCodeCommit)
class IAMUserPermissionCodeCommit(IAMUserPermission):
    repositories = FieldProperty(schemas.IIAMUserPermissionCodeCommit['repositories'])

@implementer(schemas.IIAMUserPermissionCodeBuildResource)
class IAMUserPermissionCodeBuildResource(Parent):
    codebuild = FieldProperty(schemas.IIAMUserPermissionCodeBuildResource['codebuild'])
    permission = FieldProperty(schemas.IIAMUserPermissionCodeBuildResource['permission'])
    console_access_enabled = FieldProperty(schemas.IIAMUserPermissionCodeBuildResource['console_access_enabled'])

@implementer(schemas.IIAMUserPermissionDeploymentPipelines)
class IAMUserPermissionDeploymentPipelines(IAMUserPermission):
    resources = FieldProperty(schemas.IIAMUserPermissionDeploymentPipelines['resources'])
    accounts = FieldProperty(schemas.IIAMUserPermissionDeploymentPipelines['accounts'])

@implementer(schemas.IIAMUserPermissionDeploymentPipelineResource)
class IAMUserPermissionDeploymentPipelineResource(Parent):
    pipeline = FieldProperty(schemas.IIAMUserPermissionDeploymentPipelineResource['pipeline'])
    permission = FieldProperty(schemas.IIAMUserPermissionDeploymentPipelineResource['permission'])
    console_access_enabled = FieldProperty(schemas.IIAMUserPermissionDeploymentPipelineResource['console_access_enabled'])


@implementer(schemas.IIAMUserPermissionSystemsManagerSession)
class IAMUserPermissionSystemsManagerSession(IAMUserPermission):
    resources = FieldProperty(schemas.IIAMUserPermissionSystemsManagerSession['resources'])
    accounts = FieldProperty(schemas.IIAMUserPermissionSystemsManagerSession['accounts'])

@implementer(schemas.IIAMUserPermissionCodeBuild)
class IAMUserPermissionCodeBuild(IAMUserPermission):
    resources = FieldProperty(schemas.IIAMUserPermissionCodeBuild['resources'])

@implementer(schemas.IIAMUserPermissionCustomPolicy)
class IAMUserPermissionCustomPolicy(IAMUserPermission):
    accounts = FieldProperty(schemas.IIAMUserPermissionCustomPolicy['accounts'])
    managed_policies = FieldProperty(schemas.IIAMUserPermissionCustomPolicy['managed_policies'])
    policies = FieldProperty(schemas.IIAMUserPermissionCustomPolicy['policies'])

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.policies = []
        self.managed_policies = []

@implementer(schemas.IIAMUserPermissions)
class IAMUserPermissions(Named, dict):
    pass

@implementer(schemas.IIAMUser)
class IAMUser(Named, Deployable):
    account = FieldProperty(schemas.IIAMUser['account'])
    username = FieldProperty(schemas.IIAMUser['username'])
    description = FieldProperty(schemas.IIAMUser['description'])
    console_access_enabled = FieldProperty(schemas.IIAMUser['console_access_enabled'])
    programmatic_access = FieldProperty(schemas.IIAMUser['programmatic_access'])
    permissions = FieldProperty(schemas.IIAMUser['permissions'])
    account_whitelist = FieldProperty(schemas.IIAMUser['account_whitelist'])

    def __init__(self, name, __parent__):
        super().__init__(name, __parent__)
        self.permissions = IAMUserPermissions('permissions', self)

@implementer(schemas.IIAMUsers)
class IAMUsers(Named, dict):
    pass

@implementer(schemas.IIAMRole)
class IAMRole(Resource, Role):
    account = FieldProperty(schemas.IIAMRole['account'])
    assume_role_policy = FieldProperty(schemas.IIAMRole['assume_role_policy'])
    policies = FieldProperty(schemas.IIAMRole['policies'])

@implementer(schemas.IIAMResource)
class IAMResource(Named):
    users = FieldProperty(schemas.IIAMResource['users'])


@implementer(schemas.ISSMDocuments)
class SSMDocuments(Named, dict):
    pass

@implementer(schemas.ISSMDocument)
class SSMDocument(Resource):
    locations = FieldProperty(schemas.ISSMDocument['locations'])
    content = FieldProperty(schemas.ISSMDocument['content'])
    document_type = FieldProperty(schemas.ISSMDocument['document_type'])

    def add_location(self, account_ref, region):
        "Add an account and region to locations if it does not already exist"
        for location in self.locations:
            if location.account == account_ref:
                for aws_region in location.regions:
                    if region == aws_region:
                        return
                location.regions.append(region)
                return
        # not seen in existing accounts, add new location
        location = AccountRegions(self)
        location.account = account_ref
        location.regions = [region]
        self.locations.append(location)


@implementer(schemas.ISSMResource)
class SSMResource(Named):
    name = 'ssm'
    ssm_documents = FieldProperty(schemas.ISSMResource["ssm_documents"])

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.ssm_documents = SSMDocuments('ssm_documents', self)

@implementer(schemas.IGlobalResources)
class GlobalResources(Named, dict):
    "Global Resources"

    def __init__(self, name, __parent__):
        super().__init__(name, __parent__)
        self['cloudtrail'] = CloudTrailResource('cloudtrail', self)
        self.cloudtrail = self['cloudtrail']
        self['codecommit'] = CodeCommit('codecommit', self)
        self.codecommit = self['codecommit']
        self['sns'] = SNS('sns', self)
        self.sns = self['sns']
        self['iam'] = IAMResource('iam', self)
        self.iam = self['iam']



"""
All things CloudFormation Init.
"""

from paco.models.base import Named
from paco.models import schemas
from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty
import troposphere.cloudformation


def export_attrs_as_dicts(obj, attrs):
    out = {}
    for name in attrs:
        value = getattr(obj, name, None)
        if value:
            out[name] = dict(value)
    return out

@implementer(schemas.ICloudFormationParameters)
class CloudFormationParameters(Named, dict):
    pass

@implementer(schemas.ICloudFormationConfigSets)
class CloudFormationConfigSets(Named, dict):

    def export_as_troposphere(self):
        # plain dict of list values
        return dict(self)

@implementer(schemas.ICloudFormationConfigurations)
class CloudFormationConfigurations(Named, dict):

    def export_as_troposphere(self):
        out = {}
        for key, value in self.items():
            out[key] = troposphere.cloudformation.InitConfig(
                **self[key].export_as_troposphere()
            )
        return out

@implementer(schemas.ICloudFormationInitVersionedPackageSet)
class CloudFormationInitVersionedPackageSet(dict):
    pass

@implementer(schemas.ICloudFormationInitPathOrUrlPackageSet)
class CloudFormationInitPathOrUrlPackageSet(dict):
    pass

@implementer(schemas.ICloudFormationInitPackages)
class CloudFormationInitPackages(Named):
    apt = FieldProperty(schemas.ICloudFormationInitPackages['apt'])
    msi = FieldProperty(schemas.ICloudFormationInitPackages['msi'])
    python = FieldProperty(schemas.ICloudFormationInitPackages['python'])
    rpm = FieldProperty(schemas.ICloudFormationInitPackages['rpm'])
    rubygems = FieldProperty(schemas.ICloudFormationInitPackages['rubygems'])
    yum = FieldProperty(schemas.ICloudFormationInitPackages['yum'])

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.apt = CloudFormationInitVersionedPackageSet()
        self.msi = CloudFormationInitPathOrUrlPackageSet()
        self.python = CloudFormationInitVersionedPackageSet()
        self.rpm = CloudFormationInitPathOrUrlPackageSet()
        self.rubygems = CloudFormationInitVersionedPackageSet()
        self.yum = CloudFormationInitVersionedPackageSet()

    def export_as_troposphere(self):
        return export_attrs_as_dicts(
            self,
            ('apt', 'msi', 'python', 'rpm', 'rubygems', 'yum')
        )

@implementer(schemas.ICloudFormationInitGroup)
class CloudFormationInitGroup(Named):
    gid = FieldProperty(schemas.ICloudFormationInitGroup['gid'])

    def export_as_troposphere(self):
        out = {}
        for name in ('gid'):
            value = getattr(self, name, None)
            if value != None:
                out[name] = value
        return out

@implementer(schemas.ICloudFormationInitGroups)
class CloudFormationInitGroups(Named, dict):

    def export_as_troposphere(self):
        out = {}
        for key, value in self.items():
            out[key] = self[key].export_as_troposphere()
        return out

@implementer(schemas.ICloudFormationInitUser)
class CloudFormationInitUser(Named):
    groups = FieldProperty(schemas.ICloudFormationInitUser['groups'])
    uid = FieldProperty(schemas.ICloudFormationInitUser['uid'])
    home_dir = FieldProperty(schemas.ICloudFormationInitUser['home_dir'])

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.groups = []

    def export_as_troposphere(self):
        out = {}
        for name in ('groups', 'uid', 'home_dir'):
            value = getattr(self, name, None)
            if name == 'home_dir':
                name = 'homeDir'
            if name == 'uid':
                value = str(value)
            if value != None:
                out[name] = value
        return out

@implementer(schemas.ICloudFormationInitUsers)
class CloudFormationInitUsers(Named, dict):

    def export_as_troposphere(self):
        out = {}
        for key, value in self.items():
            out[key] = self[key].export_as_troposphere()
        return out

@implementer(schemas.ICloudFormationInitSources)
class CloudFormationInitSources(Named, dict):

    def export_as_troposphere(self):
        out = {}
        for key, value in self.items():
            out[key] = value
        return out

@implementer(schemas.ICloudFormationInitFiles)
class CloudFormationInitFiles(Named, dict):

    def export_as_troposphere(self):
        out = {}
        for key, value in self.items():
            out[key] = self[key].export_as_troposphere()
        return out

@implementer(schemas.ICloudFormationInitFile)
class CloudFormationInitFile(Named):
    content_cfn_file = FieldProperty(schemas.ICloudFormationInitFile['content_cfn_file'])
    content_file = FieldProperty(schemas.ICloudFormationInitFile['content_file'])
    source = FieldProperty(schemas.ICloudFormationInitFile['source'])
    encoding = FieldProperty(schemas.ICloudFormationInitFile['encoding'])
    group = FieldProperty(schemas.ICloudFormationInitFile['group'])
    owner = FieldProperty(schemas.ICloudFormationInitFile['owner'])
    mode = FieldProperty(schemas.ICloudFormationInitFile['mode'])
    authentication = FieldProperty(schemas.ICloudFormationInitFile['authentication'])
    context = FieldProperty(schemas.ICloudFormationInitFile['context'])
    _content = None

    @property
    def content(self):
        "Return a string or a Troposphere CFN Function object"
        if self.content_file:
            return self.content_file
        elif self.content_cfn_file:
            return self.content_cfn_file
        return self._content

    @content.setter
    def content(self, value):
        self._content = value

    def export_as_troposphere(self):
        out = {}
        for name in ('content', 'source', 'encoding', 'group', 'owner', 'mode', 'authentication', 'context'):
            value = getattr(self, name, None)
            if value != None:
                out[name] = value

        return out

@implementer(schemas.ICloudFormationInitCommands)
class CloudFormationInitCommands(Named, dict):

    def export_as_troposphere(self):
        out = {}
        for key, command_obj in self.items():
            command_dict = {}
            for name in ('command', 'env', 'cwd', 'test', 'ignore_errors'):
                value = getattr(command_obj, name)
                if name == 'ignore_errors':
                    name = 'ignoreErrors'
                if value != None:
                    command_dict[name] = value
            out[key] = command_dict
        return out

@implementer(schemas.ICloudFormationInitCommand)
class CloudFormationInitCommand(Named):
    command = FieldProperty(schemas.ICloudFormationInitCommand['command'])
    env = FieldProperty(schemas.ICloudFormationInitCommand['env'])
    cwd = FieldProperty(schemas.ICloudFormationInitCommand['cwd'])
    test = FieldProperty(schemas.ICloudFormationInitCommand['test'])
    ignore_errors = FieldProperty(schemas.ICloudFormationInitCommand['ignore_errors'])

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.env = {}

@implementer(schemas.ICloudFormationInitService)
class CloudFormationInitService(Named, dict):
    ensure_running = FieldProperty(schemas.ICloudFormationInitService['ensure_running'])
    enabled = FieldProperty(schemas.ICloudFormationInitService['enabled'])
    files = FieldProperty(schemas.ICloudFormationInitService['files'])
    sources = FieldProperty(schemas.ICloudFormationInitService['sources'])
    packages = FieldProperty(schemas.ICloudFormationInitService['packages'])
    commands = FieldProperty(schemas.ICloudFormationInitService['commands'])

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.files = []
        self.packages = {}
        self.commands = []
        self.sources = []

@implementer(schemas.ICloudFormationInitServiceCollection)
class CloudFormationInitServiceCollection(Named, dict):

    def export_as_troposphere(self):
        out = {}
        for key, service_obj in self.items():
            service_dict = {}
            for name in ('ensure_running', 'enabled', 'files', 'sources', 'packages', 'commands'):
                value = getattr(service_obj, name)
                if name == 'ensure_running':
                    name = 'ensureRunning'
                if value != None:
                    service_dict[name] = value
            out[key] = service_dict
        return out

@implementer(schemas.ICloudFormationInitServices)
class CloudFormationInitServices(Named):
    sysvinit = FieldProperty(schemas.ICloudFormationInitServices['sysvinit'])
    windows = FieldProperty(schemas.ICloudFormationInitServices['windows'])

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.sysvinit = CloudFormationInitServiceCollection('sysvinit', self)
        self.windows = CloudFormationInitServiceCollection('windows', self)

    def export_as_troposphere(self):
        out = {}
        if self.sysvinit:
            out['sysvinit'] = self.sysvinit.export_as_troposphere()
        if self.windows:
            out['windows'] = self.windows.export_as_troposphere()
        return out

@implementer(schemas.ICloudFormationConfiguration)
class CloudFormationConfiguration(Named):
    packages = FieldProperty(schemas.ICloudFormationConfiguration['packages'])
    groups = FieldProperty(schemas.ICloudFormationConfiguration['groups'])
    users = FieldProperty(schemas.ICloudFormationConfiguration['users'])
    sources = FieldProperty(schemas.ICloudFormationConfiguration['sources'])
    files = FieldProperty(schemas.ICloudFormationConfiguration['files'])
    commands = FieldProperty(schemas.ICloudFormationConfiguration['commands'])
    services = FieldProperty(schemas.ICloudFormationConfiguration['services'])

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.packages = CloudFormationInitPackages('packages', self)
        self.files = CloudFormationInitFiles('files', self)
        self.commands = CloudFormationInitCommands('commands', self)
        self.services = CloudFormationInitServices('services', self)
        self.sources = CloudFormationInitSources('sources', self)
        self.groups = CloudFormationInitGroups('groups', self)
        self.users = CloudFormationInitUsers('users', self)

    def export_as_troposphere(self):
        out = {}
        for name in ('packages', 'files', 'commands', 'services', 'sources', 'groups', 'users'):
            obj = getattr(self, name, None)
            if obj:
                out[name] = obj.export_as_troposphere()
        return out

@implementer(schemas.ICloudFormationInit)
class CloudFormationInit(Named):
    config_sets = FieldProperty(schemas.ICloudFormationInit['config_sets'])
    configurations = FieldProperty(schemas.ICloudFormationInit['configurations'])
    parameters = FieldProperty(schemas.ICloudFormationInit['parameters'])

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.parameters = {}

    def export_as_troposphere(self):

        init_resource = troposphere.cloudformation.Init(
            troposphere.cloudformation.InitConfigSets(
                **self.config_sets.export_as_troposphere()
            ),
            **self.configurations.export_as_troposphere()
        )
        return init_resource


from paco.models import schemas
from paco.models.base import Named, match_allowed_paco_filenames
from zope.interface import implementer
import pkg_resources


@implementer(schemas.IServices)
class Services(Named, dict):

    enabled_services = []


def list_enabled_services(paco_home):
    """
    Return an ordered dict of Services enabled in this Paco Project.
    Sorted according to the SERVICE_INITIALIZATION_ORDER for each Service.
    """
    load_order = []
    # ToDo: 'services' should be legacy and only support 'service' ... ?
    services_dir_names = ('service', 'services')
    count = 1000 # services with no initialization order start at 1000
    for entry_point in pkg_resources.iter_entry_points('paco.services'):
        # # Legacy directory name
        # if os.path.isdir(paco_home / 'Services'):
        #     services_dir_name = 'Services'
        # services_dir = paco_home / services_dir_name
        # fname = None
        for service_dirname in services_dir_names:
            yaml_path = match_allowed_paco_filenames(paco_home, service_dirname, entry_point.name)
            if yaml_path != None:
                module = entry_point.load()
                if not hasattr(module, 'SERVICE_INITIALIZATION_ORDER'):
                    module.SERVICE_INITIALIZATION_ORDER = count
                    count += 1
                load_order.append((module.SERVICE_INITIALIZATION_ORDER, entry_point.name, module, yaml_path))
    load_order = sorted(load_order)
    entry_point_dict = {}
    for item in load_order:
        # The Service name is lower-cased
        entry_point_dict[item[1]] = {'name': item[1].lower(), 'module': item[2], 'yaml_path': item[3]}
    return entry_point_dict



# Hooks that can be registered by paco.extend
EXTEND_BASE_MODEL_HOOKS = []
CW_ALARM_ACTIONS_HOOK = None
CW_ALARM_HOOKS = []
CW_ALARM_DESCRIPTION_HOOK = None
CODESTAR_NOTIFICATION_RULE_HOOK = None
SECURITY_GROUPS_HOOKS = []


import unittest
import paco.models.exceptions
from paco.models import references


class TestReferenceClass(unittest.TestCase):

    def test_secret_base_ref(self):
        base_ref = 'paco.ref netenv.mynet.dev.eu-central-1.secrets_manager.myapp.mygroup.mydb'
        test_refs = [
            'paco.ref netenv.mynet.dev.eu-central-1.secrets_manager.myapp.mygroup.mydb',
            'paco.ref netenv.mynet.dev.eu-central-1.secrets_manager.myapp.mygroup.mydb.arn',
            'paco.ref netenv.mynet.dev.eu-central-1.secrets_manager.myapp.mygroup.mydb.myjsonfield',
            'paco.ref netenv.mynet.dev.eu-central-1.secrets_manager.myapp.mygroup.mydb.myjsonfield.arn',
        ]
        for test_ref in test_refs:
            ref_obj = references.Reference(test_ref)
            base_ref_obj = ref_obj.secret_base_ref()
            assert base_ref_obj.raw == base_ref
        bogus_ref = 'paco.ref netenv.mynet.dev.eu-central-1.applications.myapp'
        ref_obj = references.Reference(bogus_ref)
        with self.assertRaises(paco.models.exceptions.InvalidPacoReference):
            ref_obj.secret_base_ref()




"""
AWS Backup models
"""

from paco.models.base import Named, Parent, Resource
from paco.models import schemas
from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty

@implementer(schemas.IBackupPlanCopyActionResourceType)
class BackupPlanCopyActionResourceType(Parent):
    destination_vault = FieldProperty(schemas.IBackupPlanCopyActionResourceType['destination_vault'])
    lifecycle_delete_after_days = FieldProperty(schemas.IBackupPlanCopyActionResourceType['lifecycle_delete_after_days'])
    lifecycle_move_to_cold_storage_after_days = FieldProperty(schemas.IBackupPlanCopyActionResourceType['lifecycle_move_to_cold_storage_after_days'])

@implementer(schemas.IBackupPlanRule)
class BackupPlanRule(Named):
    lifecycle_delete_after_days = FieldProperty(schemas.IBackupPlanRule['lifecycle_delete_after_days'])
    lifecycle_move_to_cold_storage_after_days = FieldProperty(schemas.IBackupPlanRule['lifecycle_move_to_cold_storage_after_days'])
    schedule_expression = FieldProperty(schemas.IBackupPlanRule['schedule_expression'])
    copy_actions = FieldProperty(schemas.IBackupPlanRule['copy_actions'])

@implementer(schemas.IBackupSelectionConditionResourceType)
class BackupSelectionConditionResourceType(Parent):
    condition_type = FieldProperty(schemas.IBackupSelectionConditionResourceType['condition_type'])
    condition_key = FieldProperty(schemas.IBackupSelectionConditionResourceType['condition_key'])
    condition_value = FieldProperty(schemas.IBackupSelectionConditionResourceType['condition_value'])


@implementer(schemas.IBackupPlanSelection)
class BackupPlanSelection(Parent):
    title = FieldProperty(schemas.IBackupPlanSelection['title'])
    tags = FieldProperty(schemas.IBackupPlanSelection['tags'])
    resources = FieldProperty(schemas.IBackupPlanSelection['resources'])

@implementer(schemas.IBackupPlan)
class BackupPlan(Resource):
    plan_rules = FieldProperty(schemas.IBackupPlan['plan_rules'])
    selections = FieldProperty(schemas.IBackupPlan['selections'])

@implementer(schemas.IBackupPlans)
class BackupPlans(Named, dict):
    pass

@implementer(schemas.IBackupVault)
class BackupVault(Resource):
    notification_events = FieldProperty(schemas.IBackupVault['notification_events'])
    notification_group = FieldProperty(schemas.IBackupVault['notification_group'])
    plans = FieldProperty(schemas.IBackupVault['plans'])

@implementer(schemas.IBackupVaults)
class BackupVaults(Named, dict):
    pass

